/***
 * OW2 FraSCAti Assembly Factory
 * Copyright (C) 2009 INRIA, USTL
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: frascati@ow2.org 
 *
 * Author: Philippe Merle
 * 
 * Contributors: Nicolas Dolet, Christophe Demarey
 */

package org.ow2.frascati.factory.core.instance.implementation.script;

import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.URL;
import javax.script.Invocable;
import javax.script.ScriptException;
import javax.script.ScriptEngine;
import javax.script.ScriptEngineFactory;
import javax.script.ScriptEngineManager;
import org.eclipse.stp.sca.Implementation;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.api.control.ContentController;
import org.objectweb.fractal.api.factory.GenericFactory;
import org.objectweb.fractal.api.type.ComponentType;
import org.objectweb.fractal.api.type.InterfaceType;
import org.objectweb.fractal.api.type.TypeFactory;
import org.objectweb.fractal.fraclet.annotations.Requires;
import org.objectweb.fractal.util.Fractal;
import org.ow2.frascati.factory.core.instance.NotSupportedException;
import org.ow2.frascati.factory.core.instance.implementation.ScaImplementation;
import org.ow2.frascati.factory.core.instance.implementation.ImplementationException;
import org.ow2.frascati.factory.core.instance.runtime.RuntimeFactory;
import org.ow2.frascati.factory.core.parser.util.ScaModelHelper;
import org.eclipse.stp.sca.domainmodel.frascati.FrascatiPackage;
import org.eclipse.stp.sca.domainmodel.frascati.ScriptImplementation;
import org.ow2.frascati.factory.core.instance.InstantiationException;

/**
 * Create FraSCAti component instances implemented by scripts.
 * 
 * @author <a href="mailto:Philippe.Merle@inria.fr">Philippe Merle</a>
 *
 * @version 1.1
 */

@org.objectweb.fractal.fraclet.annotations.Component
public class ScaScriptImpl implements ScaImplementation {

  // ======================================================================
  //
  // Internal state.
  //
  // ======================================================================

  /**
   * The implementation type calculated with Object Eclass.
   */
  private static String IMPLEMENTATION_TYPE = ScaModelHelper
      .getID(FrascatiPackage.Literals.SCRIPT_IMPLEMENTATION);

  /**
   * Logger.
   */
  private java.util.logging.Logger log = java.util.logging.Logger
      .getLogger(this.getClass().getCanonicalName());

  /**
   * The script engine manager.
   */
  private ScriptEngineManager scriptEngineManager;

  /**
   * The runtime to create Fractal components.
   */
  @Requires(name = "runtime")
  private RuntimeFactory runtime;

  // ======================================================================
  //
  // Internal methods.
  //
  // ======================================================================

  /**
   * Initialize the script engine manager.
   */
  protected void initializeScriptEngineManager() {
    // Create the script engine manager with the runtime classloader.
    // Let's note that this can't be done before (e.g., at the instantiation time
    // of this component) because the runtime classloader could not be initialized too.
    // Christophe pointed this when applying the FraSCAti MOJO plugin to execute
    // 'mvn -Prun' on examples/helloworld-script/
    scriptEngineManager = new ScriptEngineManager(runtime.getClassLoader());

    // Log currently supported scripting engines.
    StringBuffer buffer = new StringBuffer();
    buffer.append("FraSCAti - Supported scripting engines:\n");
    for(ScriptEngineFactory factory : scriptEngineManager.getEngineFactories()) {
      buffer.append("* ");
      buffer.append(factory.getEngineName());
      buffer.append(' ');
      buffer.append(factory.getEngineVersion());
      buffer.append(' ');
      buffer.append(factory.getExtensions());
      buffer.append(' ');
      buffer.append(factory.getNames());
      buffer.append(' ');
      buffer.append(factory.getLanguageName());
      buffer.append(' ');
      buffer.append(factory.getLanguageVersion());
      buffer.append('\n');
    }
    log.info(buffer.toString());
  }

  // ======================================================================
  //
  // Implementation of
  // org.ow2.frascati.factory.core.instance.implementation.ScaImplementation
  //
  // ======================================================================

  /**
   * (non-Javadoc)
   * 
   * @see ScaImplementation#getImplementationID()
   */
  public String getImplementationID() {
    return IMPLEMENTATION_TYPE;
  }

  /**
   * (non-Javadoc)
   * 
   * @see ScaImplementation#create(Implementation,ComponentType)
   */
  public Component create(final Implementation implementation,
      final ComponentType componentType)
    throws ImplementationException, NotSupportedException
  {
    // Initialize the script engine manager if not already done.
    // At this moment the FraSCAti assembly factory runtime classloader
    // is initialized.
    //
    // TODO: Perhaps the script engine manager should be instantiated
    // each time a ScriptEngine is instantiated in order to be sure that
    // all script engines present in the runtime classloader are useable.
    // For instance, we could imagine a reconfiguration scenario where
    // we dynamically add a new script engine at runtime.
    if(scriptEngineManager == null) {
      initializeScriptEngineManager();
    }

	// Implementation is a script one.
    ScriptImplementation scriptImplementation = (ScriptImplementation)implementation;
    // Get the script.
    String script = scriptImplementation.getScript();

    log.info("Create an SCA component for <frascati:implementation.script script=\""
        + script + "\" language=\"" + scriptImplementation.getLanguage() + "\"> and the Fractal component type "
        + componentType.toString());

    // Obtain an input stream reader to the script.
    // TODO: Could 'script' contain a url or filename, and not just a resource in the classpath?
    URL scriptUrl = runtime.getClassLoader().getResource(script);
    if(scriptUrl == null) {
        throw new ImplementationException("Script '" + script + "' not found");
    }
    InputStream scriptInputStream = null;
    try {
    	scriptInputStream = scriptUrl.openStream();
    } catch (IOException ioe) {
        throw new ImplementationException("Script '" + script + "' not found", ioe);
    }
    InputStreamReader scriptReader = new InputStreamReader(scriptInputStream);

    // Obtain a scripting engine for evaluating the script.
    // TODO: Use scriptImplementation.getLanguage() if not null
    String scriptExtension = script.substring(script.lastIndexOf('.') + 1);
    ScriptEngine scriptEngine = scriptEngineManager.getEngineByExtension(scriptExtension);
    if(scriptEngine == null) {
        throw new ImplementationException("ScriptEngine for '" + script + "' not found");
    }

    // Evaluate the script.
    try {
        scriptEngine.eval(scriptReader);
    } catch(ScriptException se) {
        throw new ImplementationException("Error when evaluating '" + script + "'", se);
    }
    Invocable invocableScriptEngine = (Invocable)scriptEngine;

    Component component = null;
    try {
        // Create the Fractal component via the generic factory.
        component = runtime.newInstance(componentType, "scaComposite", null);
    } catch (InstantiationException ie) {
        // This should not happen!
        throw new ImplementationException("Can't create the SCA component!", ie);
    }
    BindingController bindingController = null;
    ContentController contentController = null;
    try {
    	bindingController = Fractal.getBindingController(component);
        contentController = Fractal.getContentController(component);
    } catch (NoSuchInterfaceException nsie) {
        // This should not happen!
        throw new ImplementationException("Internal Fractal error!", nsie);
    }

    // Traverse all the interface types of the component type.
    for (InterfaceType interfaceType : componentType.getFcInterfaceTypes()) {
        String interfaceName = interfaceType.getFcItfName();
        if (!interfaceType.isFcClientItf()) {
            // When this is a server interface type.
            String interfaceSignature = interfaceType.getFcItfSignature();
            Class<?> interfaze;
            try {
                interfaze = runtime.getClassLoader().loadClass(interfaceSignature);
            } catch(ClassNotFoundException cnfe) {
                throw new ImplementationException("Interface '" + interfaceSignature + "' not found", cnfe);
            }
            // Get a typed proxy to the script.
            Object proxy = invocableScriptEngine.getInterface(interfaze);
            try {
                // Bind the Fractal component to the typed proxy.
                bindingController.bindFc(interfaceName, proxy);
            } catch (Exception exc) {
                // This should not happen!
                throw new ImplementationException("Internal Fractal error!", exc);
            }
        } else {
            // When this is a client interface type.
            try {
                Object itf = contentController.getFcInternalInterface(interfaceName);
                // Put the interface as a variable of the scripting engine.
                scriptEngine.put(interfaceName, itf);
            } catch(NoSuchInterfaceException nsie) {
                 // This should not happen!
                 throw new ImplementationException("Internal Fractal error!", nsie);
            }
        }
    }

    // Create a primitive component encapsulating the scripting engine.
    try {
    	ScriptEngineComponent scriptEngineContent = new ScriptEngineComponent(component, scriptEngine);

        // Here use Julia not Tinfi.
    	Component bootstrap = Fractal.getBootstrapComponent();
		TypeFactory typeFactory = Fractal.getTypeFactory(bootstrap);
		GenericFactory genericFactory = Fractal.getGenericFactory(bootstrap);

        ComponentType scriptEngineType = typeFactory.createFcType(new InterfaceType[] {
            // type for attribute-controller
            typeFactory.createFcItfType("attribute-controller",
                ScriptEngineAttributes.class.getCanonicalName(),
                false, false, false) });
        Component scriptEngineComponent = genericFactory.newFcInstance(scriptEngineType,
                "primitive", scriptEngineContent);

        // add the scripting engine component as a sub component of the SCA composite.
        contentController.addFcSubComponent(scriptEngineComponent);
    } catch(Exception exc) {
        // This should not happen!
        throw new ImplementationException("Internal Fractal error!", exc);
    }

    // Return the Fractal component.
    return component;
  }
}
