/***
 * OW2 FraSCAti Assembly Factory
 * Copyright (C) 2008-2009 INRIA, USTL
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: frascati@ow2.org 
 *
 * Author: Damien Fournier
 */

package org.ow2.frascati.factory.core.parser.util;

import java.io.IOException;

import org.eclipse.emf.common.util.EList;
import org.eclipse.stp.sca.Composite;
import org.eclipse.stp.sca.Include;
import org.ow2.frascati.factory.core.instance.runtime.RuntimeFactory;
import org.ow2.frascati.factory.core.parser.Sca2Ecore;

/**
 * Add to current Composite the components, services, references, properties and
 * wires defined in included composite
 * 
 * @author <a href="mailto:damien.fournier@inria.fr">Damien Fournier</a>
 * @version $Revision$
 */

public class IncludeResolver implements
    org.eclipse.stp.sca.introspection.Resolver {

  /**
   * parser instance used for recursive calls
   */

  protected Sca2Ecore parser;

  /**
   * Reference on runtime component
   */

  protected RuntimeFactory runtime;

  /**
   * Initialize the include Resolver with a parser instance
   * 
   * @param parser
   */

  public IncludeResolver(Sca2Ecore parser, RuntimeFactory runtime) {
    this.parser = parser;
    this.runtime = runtime;
  }

  /**
   * Logger for this class
   */

  java.util.logging.Logger log = java.util.logging.Logger
      .getLogger(JavaInterfaceResolver.class.getCanonicalName());

  /**
   * Resolve <include> for the provided Composite model element
   * 
   * @param composite
   * @return - a Composite with <include> resolved
   * @throws IOException
   */

  public Composite resolve(Composite composite) {

    EList<Include> list = composite.getInclude();

    if (list.isEmpty()) {
      // Include not defined in main composite file : nothing to do
      return composite;
    } else {
      // Included composite had been defined
      for (Include include : list) {

        String includeComposite = getIncludeURI(include);

        try {

          composite = resolveIncluded(composite, parser.parseComposite(runtime
              .getCompositeResource(includeComposite)));

        } catch (IOException e) {
          log.severe("Included File " + includeComposite
              + " could not be loaded");
        }
      }

    }

    return composite;

  }

  /**
   * Internal recursive call
   * 
   * @param composite
   *          - current composite
   * @param includedComposite
   *          - composite found in the <include> element
   * @return - Merged composite
   * @throws IOException
   */

  private Composite resolveIncluded(Composite composite,
      Composite includedComposite) throws IOException {

    EList<Include> list = includedComposite.getInclude();

    if (!list.isEmpty()) {
      for (Include include : list) {

        if (include.getName() != null)
          try {
            includedComposite = resolveIncluded(includedComposite, parser
                .parseComposite(runtime.getCompositeResource(getIncludeURI(include))));
          } catch (IOException e) {
            log.severe("Included File " + include.getName()
                + " could not be loaded");
          }
        else
          log.severe("Could not get name for included composite, composite include will be ignored");

      }
    }

    // Include is actually not supported with PolicySets
    // composite.getPolicySets().addAll(includedComposite.getPolicySets());

    composite.getComponent().addAll(includedComposite.getComponent());
    composite.getProperty().addAll(includedComposite.getProperty());
    composite.getService().addAll(includedComposite.getService());
    composite.getReference().addAll(includedComposite.getReference());
    composite.getWire().addAll(includedComposite.getWire());

    return composite;

  }

  /**
   * Retrieve URI for include attribute
   * 
   * @param QName
   *          object referencing the included composite
   * @return URI of the included composite
   */
  private String getIncludeURI(Include include) {

    // retrieve URI for included composite
    String includeURI = include.getName().getLocalPart();

    // Retrieve file URI if prefix is present
    String prefix = include.getName().getPrefix();

    if (prefix != null && !prefix.equals("")) {
      String prefixUri = include.getName().getNamespaceURI();

      // Check if namespace uri ends with "/"
      if (!prefixUri.endsWith("/"))
        prefixUri = prefixUri + "/";

      // Concatenate URI string
      includeURI = prefixUri + includeURI;
    }

    return includeURI;
  }

}
