/***
 * OW2 FraSCAti explorer
 * Copyright (C) 2009 INRIA, USTL
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: frascati@ow2.org 
 *
 * Author: Christophe Demarey
 * 
 * Contributor(s):
 */
package org.ow2.frascati.explorer;

import java.util.Map;

import org.objectweb.fractal.adl.ContextMap;
import org.objectweb.fractal.adl.Factory;
import org.objectweb.fractal.adl.FactoryFactory;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.util.Fractal;
import org.ow2.frascati.factory.core.domain.DomainException;
import org.ow2.frascati.factory.runtime.domain.api.Domain;

/**
 * FraSCAti Explorer launcher.
 * It allows to:
 * <ul>
 * <li>instanciate the FraSCAti runtime,</li>
 * <li>instanciate the explorer architecture,</li>
 * <li>eventually load SCA composites,</li>
 * <li>and show the FraSCAti explorer GUI.</li>
 * 
 * @author Christophe Demarey.
 */
public class FrascatiExplorerLauncher {
  /** Default Explorer ADL file */
  private static final String DEFAULT_EXPLORER_ADL = "org.ow2.frascati.explorer.FrascatiExplorer";
  /** The explorer instance */
  protected Component explorer;
  
  /**
   * Default Constructor.
   */
  public FrascatiExplorerLauncher() {
    String adlFile = System.getProperty("frascati.explorer.adl");
    if (adlFile != null)
        this.explorer = loadArchitecture(adlFile);
    else
        this.explorer = loadArchitecture(FrascatiExplorerLauncher.DEFAULT_EXPLORER_ADL);
  }

  /**
   * Load the Explorer architecture with the help of the ADL factory.
   * 
   * @return the Explorer composite.
   */
  private static Component loadArchitecture(final String explorerAdl) {
    Component explorer = null;  
    Factory factory = null;
    Map<Object, Object> context = new ContextMap();
    
    context.put("title", "FraSCAti Explorer");
    try {
      factory = FactoryFactory.getFactory(FactoryFactory.FRACTAL_BACKEND);
      explorer = (Component) factory.newComponent(explorerAdl, context);
      Fractal.getLifeCycleController(explorer).startFc();
    } catch (Exception e) {
      e.printStackTrace();
      System.exit(1);
    }     
    return explorer;
  }
  
  /**
   * Load, instanciate and start SCA composite in the runtime.
   *  
   * @param composites - SCA composite file names.
   */
  protected void loadComposites(String[] composites) {
    Domain domain = null;
    
    try {
      domain = (Domain) explorer.getFcInterface("domain");

      for (String composite : composites)
        domain.getComposite(composite);
    } catch (DomainException e1) {
      e1.printStackTrace();
    } catch (NoSuchInterfaceException e) {
      e.printStackTrace();
      System.exit(1);
    }    
  }
  
  /**
   * Show the FraSCAti Explorer GUI.
   */
  protected void showGui() {
    try {
      Runnable r = (Runnable) explorer.getFcInterface("r");
      r.run();
    } catch (Exception e) {
      e.printStackTrace();
    }    
  }
  
  /**
   * The bootstrap method.
   * 
   * @param args - program arguments.
   */
  public static void main(String[] args) {
    FrascatiExplorerLauncher launcher = new FrascatiExplorerLauncher();
    
    launcher.loadComposites(args);   
    launcher.showGui();
  }

}
