/***
 * OW2 FraSCAti fscript
 * Copyright (C) 2009 INRIA, USTL
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: frascati@ow2.org
 *
 * Author: Christophe Demarey
 *
 * Contributor(s):
 */

package org.ow2.frascati.fscript;

import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.objectweb.fractal.adl.ContextMap;
import org.objectweb.fractal.adl.Factory;
import org.objectweb.fractal.adl.FactoryFactory;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.fscript.FScriptEngine;
import org.objectweb.fractal.fscript.FScriptException;
import org.objectweb.fractal.fscript.model.Node;
import org.objectweb.fractal.util.Fractal;
import org.ow2.frascati.fscript.model.NodeFactory;

/**
 * Utility class used to obtain the FraSCAti Script engine.
 * 
 * @author Christophe Demarey.
 */
public class FraSCAtiScript {
  /** The default FraSCAti Script architecture definition. */
  public static final String DEFAULT_ADL_DEF = "org.ow2.frascati.fscript.FraSCAtiScript";
    
  /** The FraSCAti script assembly. */
  private Component fscript = null;
  
  /** The singleton */
  private static FraSCAtiScript singleton = null;
  
  /** The logger */
  private static Logger log = Logger.getLogger("org.ow2.frascati.fscript.FraSCAtiScript");
  
  /**
   * Create a FraSCAtiScript singleton with the specified implementation.
   * 
   * @param adl - The name of the fractal ADL definition to use to instantiate FraSCAtiScript.
   */
  public static void createSingleton(final String adl) {
      if (singleton == null) singleton = new FraSCAtiScript(adl);
  }
  
  /**
   * Get the FraSCAtiScript singleton.
   * 
   * @return the FraSCAtiScript singleton.
   */
  public static FraSCAtiScript getSingleton() {
      return singleton;
  }
  
  /**
   * Instanciates a FraSCAtiScript assembly from the given ADL definition.
   * 
   * @param adl The architecture to instantiate. 
   */
  private FraSCAtiScript(final String adl) {
      Map<Object, Object> context = new ContextMap();
    
      context.put("title", "FraSCAti Explorer");
      this.fscript = loadArchitecture(adl, context);
  }
  
  /**
   * Load the FraSCAtiScript architecture with the help of the ADL factory.
   * 
   * @param adl the ADL definition to load.
   * 
   * @return the root composite.
   */
  private static Component loadArchitecture(final String adl, final Map<Object, Object> context) {
    Component root = null;  
    Factory factory = null;
    try {
      factory = FactoryFactory.getFactory(FactoryFactory.FRACTAL_BACKEND);
      log.info("Loading FraSCAti Script: " + adl + " definition.");
      root = (Component) factory.newComponent(adl, context);
      Fractal.getLifeCycleController(root).startFc();
    } catch (Exception e) {
      log.severe("Cannot instantiate " + adl +"!");
      e.printStackTrace();
      System.exit(1);
    }     
    return root;
  }
  
  /**
   * Get the {@link FScriptEngine}.
   * 
   * @return the engine's <code>FScriptEngine</code> interface
   */
  public FScriptEngine getFraSCAtiScriptEngine() {
      try {
          return (FScriptEngine) fscript.getFcInterface("engine");
      } catch (NoSuchInterfaceException e) {
          log.severe("Cannot found the 'engine' interface on the FraSCAti Script component.");
          return null;
      }
  }

  /**
   * Get the {@link NodeFactory}.
   * 
   * @return the engine's <code>NodeFactory</code> interface
   */
  public NodeFactory getNodeFactory() {
      try {
          return (NodeFactory) fscript.getFcInterface("sca-node-factory");
      } catch (NoSuchInterfaceException e) {
          log.severe("Cannot found the 'sca-node-factory' interface on the FraSCAti Script component.");
          return null;
      }
  }
  
  /**
   * Get the root component/assembly of the FraSCAti Script architecture.
   * 
   * @return the FraSCAti Script root component.
   */
  public Component getRootComponent() {
      return this.fscript;
  }
  
  /**
   * Update FraSCAti SCript context by setting the varName variable 
   * with the given component.
   * 
   * @param varName The variable name to set.
   * @param comp A fractal component that will be registered in the 'varName' variable.
   */
  public void updateContext(String varName, Component comp) {
      NodeFactory nf = getNodeFactory();
      Node rootNode = nf.createScaComponentNode(comp);
      FScriptEngine fscript = getFraSCAtiScriptEngine();
      
      fscript.setGlobalVariable(varName, rootNode);
  }

  /**
   * Execute a FraSCAti Script command.
   * 
   * @param cmd The FraSCAti Script command to execute.
   * @return the result of the command.
   */
  public Object execute(String cmd) 
    throws FScriptException {
      FScriptEngine fscript = getFraSCAtiScriptEngine();
      
      return fscript.execute(cmd); 
  }
}
