/***
 * OW2 FraSCAti Tinfi
 * Copyright (C) 2007-2009 INRIA, USTL
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: frascati@ow2.org
 *
 * Author: Lionel Seinturier
 */

package org.ow2.frascati.tinfi.juliac;

import java.io.IOException;

import org.objectweb.fractal.api.control.AttributeController;
import org.objectweb.fractal.api.type.InterfaceType;
import org.objectweb.fractal.julia.control.attribute.CloneableAttributeController;
import org.objectweb.fractal.julia.type.BasicInterfaceType;
import org.objectweb.fractal.juliac.InitializerClassGenerator;
import org.objectweb.fractal.juliac.Juliac;
import org.objectweb.fractal.juliac.JuliacConfig;
import org.objectweb.fractal.juliac.proxy.AttributeControllerClassGenerator;
import org.objectweb.fractal.juliac.proxy.InterfaceImplementationClassGenerator;
import org.objectweb.fractal.juliac.proxy.ProxyClassGenerator;
import org.objectweb.fractal.juliac.ucf.UnifiedClass;
import org.osoa.sca.annotations.Callback;

/**
 * This class generates the source code associated to Fractal components.
 * The membrane implementation and the initializer implementation are generated.
 * 
 * The content, the interceptors and the controllers are kept in separate
 * classes and the controllers are implemented with components.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 */
public class FCCompCtrlSourceCodeGenerator
extends org.objectweb.fractal.juliac.opt.comp.FCCompCtrlSourceCodeGenerator {
    
    public FCCompCtrlSourceCodeGenerator() {
        super();        
    }

    
    // -----------------------------------------------------------------------
    // Implementation of the OptLevelSourceCodeGenerator interface
    // -----------------------------------------------------------------------

    /**
     * Initialize this generator.
     * 
     * The initialization process is kept separated from the constructor in
     * order to be able to customize the julia.config system property.
     * 
     * @param jc  the instance of Juliac using this source code generator
     */
    @Override
    public void init( Juliac jc ) throws IOException {
        super.init(jc);

        /*
         * Register component interceptor code generators.
         */
        JuliacConfig jconf = jc.getJuliacConfig();
        jconf.putInterceptorSourceCodeGenerator(
            "org.ow2.frascati.tinfi.asm.SCAContentInterceptorCodeGenerator",
            SCAContentInterceptorSourceCodeGenerator.class.getName() );
    }

    
    // -----------------------------------------------------------------------
    // Overriding methods from org.objectweb.fractal.juliac.opt.comp.FCCompCtrlSourceCodeGenerator
    // -----------------------------------------------------------------------

    /**
     * Return the source code generators for the initializer classes associated
     * with this component source code generator.
     */
    @Override
    protected InitializerClassGenerator[] getInitializerSourceCodeGenerators() {
        return
            new InitializerClassGenerator[]{
                new InitializerCompCtrlSourceCodeGenerator()
        };
    }

    /**
     * Generate the source code of component interfaces associated with the
     * specified interface type.
     */
    @Override
    protected void generateInterfaceImpl( InterfaceType it, String ctrlDesc )
    throws IOException {
        
        String signature = it.getFcItfSignature();
        UnifiedClass cl = jc.create(signature);        
        UnifiedClass ac = jc.create(AttributeController.class.getName());
        UnifiedClass cac = jc.create(CloneableAttributeController.class.getName());
        
        /*
         * If the interface is a business attribute controller interface,
         * generate an implementation which provides a cloneFcAttributes
         * method.
         * 
         * The second expression in the && skips cloneable attribute control
         * interfaces which are provided by parametric membranes.
         */
        if( ac.isAssignableFrom(cl) && ! cac.isAssignableFrom(cl) ) {
            ProxyClassGenerator ascg =
                new AttributeControllerClassGenerator(it,null,false,null);
            jc.generateSourceCode(ascg);
        }
        
        InterfaceSourceCodeGeneratorFactory iscgf =
            new InterfaceSourceCodeGeneratorFactory(it,ctrlDesc,null,jc);
        
        if( iscgf.isSCACompBusinessInterface() ) {
           
            /*
             * Generate SCA specific code interface implementations.
             */

            // Server (input) interfaces
            ProxyClassGenerator scg = iscgf.getInterfaceSourceCodeGenerator();
            jc.generateSourceCode(scg);
                
            // ServiceReference implementation
            scg = new ServiceReferenceClassGenerator(it,null,false,jc);
            jc.generateSourceCode(scg);
            
            // CallableReference and input interface implementations for callbacks
            Callback annot = cl.getAnnotation(Callback.class);
            if( annot != null ) {
                Class<?> cbcl = annot.value();
                InterfaceType cbit = new BasicInterfaceType("callback",cbcl.getName(),false,false,false);
                scg = new CallableReferenceClassGenerator(cbit,null,false,jc);
                jc.generateSourceCode(scg);
                
                scg = new CallBackInterfaceClassGenerator(cbit,null,false,jc);
                jc.generateSourceCode(scg);
            }
        }
        else {
            
            /*
             * Interfaces of non-SCA components.
             */

            ProxyClassGenerator sfg = 
                new InterfaceImplementationClassGenerator(it,null,false,jc);
            jc.generateSourceCode(sfg);
        }
    }
}
