/***
 * OW2 FraSCAti Tinfi
 * Copyright (C) 2007-2009 INRIA, USTL
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: frascati@ow2.org
 *
 * Author: Lionel Seinturier
 * Contributor: Philippe Merle
 */

package org.ow2.frascati.tinfi;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.Interface;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.factory.InstantiationException;
import org.objectweb.fractal.julia.InitializationContext;
import org.objectweb.fractal.julia.Interceptor;
import org.osoa.sca.RequestContext;
import org.ow2.frascati.tinfi.control.component.ReconfigurableComponentContext;
import org.ow2.frascati.tinfi.control.content.ContentInstantiationException;
import org.ow2.frascati.tinfi.control.content.SCAContentController;
import org.ow2.frascati.tinfi.control.intent.IntentHandler;
import org.ow2.frascati.tinfi.osoa.RequestContextImpl;

/**
 * The root class for Tinfi component interceptors.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 * @author Philippe Merle <Philippe.Merle@inria.fr>
 */
public abstract class TinfiComponentInterceptor<T> implements Interceptor {

    // -----------------------------------------------------------------------
    // Implementation of the Controller interface
    // -----------------------------------------------------------------------
    
    public void initFcController( InitializationContext ic )
    throws InstantiationException {
        
        Interceptor owner = (Interceptor) ic.getInterface("component");
        
        /*
         * Skip the intent interceptor which is installed on all control
         * interfaces, including "component".
         */
        this.owner = (Component) owner.getFcItfDelegate();
    }
    
    
    // -----------------------------------------------------------------------
    // Implementation of the Interceptor interface
    // -----------------------------------------------------------------------
    
    public void setFcItfDelegate( Object impl ) {
        /*
         * The impl parameter is not declared to be of type T since the compiler
         * in this case complains that the method does not override
         * setFcItfDelegate from the super class.
         */
        this.impl = (T) impl;
    }
    public T getFcItfDelegate() { return impl; }
    protected T impl;

    
    // -----------------------------------------------------------------------
    // Implementation of the Controller interface
    // -----------------------------------------------------------------------
    
    @Override
    public abstract Object clone();
    
    /**
     * Initialize a clone instance. This method is called by generated
     * subclasses.
     */
    protected void initFcClone( TinfiComponentInterceptor<T> clone ) {
        clone.intentHandlersMap = intentHandlersMap;
        clone.owner = owner;
        clone.itf = itf;
    }

    private Component owner;
    private ReconfigurableComponentContext compctx;
    private SCAContentController cc;
    private Interface itf;

    
    // -----------------------------------------------------------------------
    // Methods for managing the reference between the interceptor and its
    // associated interface
    // -----------------------------------------------------------------------

    public void setFcItf( Interface itf ) {
        this.itf = itf;
    }
    
    public Interface getFcItf() {
        return itf;
    }
    
    
    // -----------------------------------------------------------------------
    // Management of intent handlers
    // -----------------------------------------------------------------------
    
    /*
     * The Fc string in method names has been removed to avoid name clashes when
     * adding intent handlers on the SCAIntentController interface.
     */
    
    /**
     * The map of intent handlers associated with each component method managed
     * by this interceptor. This map is queried by generated subclasses.
     * 
     * @since 0.4
     */
    protected Map<Method,List<IntentHandler>> intentHandlersMap;
    
    /**
     * The array of component methods managed by this interceptor.
     * 
     * @since 1.0
     */
    protected Method[] methods;
        
    /**
     * Initialize the map of intent handlers by adding an empty entry for each
     * specified method. 
     */
    protected void initIntentHandlersMap( Method[] methods ) {
        this.methods = methods;
        intentHandlersMap = new HashMap<Method,List<IntentHandler>>();
        for (Method method : methods) {
            List<IntentHandler> handlers = new ArrayList<IntentHandler>();
            intentHandlersMap.put(method,handlers);
        }
    }
    
    /**
     * Add the specified intent handler for all component methods managed by
     * this interceptor.
     */
    public void addIntentHandler( IntentHandler handler ) {
        for (Map.Entry<Method,List<IntentHandler>> entry : intentHandlersMap.entrySet()) {
            List<IntentHandler> handlers = entry.getValue();
            handlers.add(handler);
        }
    }
    
    /**
     * Add the specified intent handler for the specified method.
     * 
     * @throws NoSuchMethodException
     *      if the specified method is not a component method managed by this
     *      interceptor
     */
    public void addIntentHandler( IntentHandler handler, Method method )
    throws NoSuchMethodException {
        
        if( ! intentHandlersMap.containsKey(method) ) {
            throw new NoSuchMethodException();
        }
        
        List<IntentHandler> handlers = intentHandlersMap.get(method);
        handlers.add(handler);
    }
    
    /**
     * Return the array of component methods managed by this interceptor.
     * 
     * @since 1.0
     */
    public Method[] getMethods() {
        return methods;
    }
    
    /**
     * Return the list of all intent handlers managed by this interceptor.
     * 
     * @since 0.4.4
     */
    public List<IntentHandler> listIntentHandler() {
        List<IntentHandler> all = new ArrayList<IntentHandler>();
        for (Map.Entry<Method,List<IntentHandler>> entry : intentHandlersMap.entrySet()) {
            List<IntentHandler> handlers = entry.getValue();
            all.addAll(handlers);
        }
        return all;
    }
    
    /**
     * Return the list of all intent handlers managed by this interceptor for
     * the specified method.
     * 
     * @throws NoSuchMethodException
     *      if the specified method is not a component method managed by this
     *      interceptor
     * @since 0.4.4
     */
    public List<IntentHandler> listIntentHandler( Method method )
    throws NoSuchMethodException {

        if( ! intentHandlersMap.containsKey(method) ) {
            throw new NoSuchMethodException();
        }
        
        List<IntentHandler> handlers = intentHandlersMap.get(method);
        List<IntentHandler> all = new ArrayList<IntentHandler>();
        all.addAll(handlers);
        return all;
    }
    
    /**
     * Remove the specified intent handler for all component methods managed by
     * this interceptor.
     */
    public void removeIntentHandler( IntentHandler handler ) {
        for (Map.Entry<Method,List<IntentHandler>> entry : intentHandlersMap.entrySet()) {
            List<IntentHandler> handlers = entry.getValue();
            handlers.remove(handler);
        }
    }
    
    /**
     * Remove the specified intent handler for the specified method.
     * 
     * @throws NoSuchMethodException
     *      if the specified method is not a component method managed by this
     *      interceptor
     */
    public void removeIntentHandler( IntentHandler handler, Method method )
    throws NoSuchMethodException {
        
        if( ! intentHandlersMap.containsKey(method) ) {
            throw new NoSuchMethodException();
        }
        
        List<IntentHandler> handlers = intentHandlersMap.get(method);
        handlers.remove(handler);
    }
    
    
    // -----------------------------------------------------------------------
    // Request context stack and content instance management.
    // These methods are used by generated subclasses.
    // -----------------------------------------------------------------------
    
    /**
     * Push a new request context on the thread-local stack of request contexts
     * and return the content instance corresponding to the current scope
     * policy.
     * 
     * @since 1.1.1
     */
    protected T pushFcAndGet(
        String serviceName, Class<T> businessInterface, T service ) {
        
        // Push a new request context
        SCAContentController cc = getFcSCAContentCtrlItf();
        RequestContext rc =
            new RequestContextImpl<T>(serviceName,businessInterface,service);
        cc.pushRequestContext(rc);        

        // Get a new content instance
        try {
            Object content = cc.getFcContent();
            return (T) content;
        }
        catch( ContentInstantiationException e ) {
            throw new TinfiRuntimeException(e);
        }
    }
    
    /**
     * Release the specified content instance previously retrieved with {@link
     * #getContent()} and pop the request context pushed by {@link
     * TinfiComponentInterceptor#pushFcRequestContext(String, Class, Object)}.
     * 
     * @param content  the content instance which is released
     * @param isEndMethod
     *      <code>true</code> if the method which releases the content instance
     *      is annotated with {@link org.osoa.sca.annotations.EndsConversation}
     * @since 1.1.1
     */
    protected void releaseFcAndPop( Object content, boolean isEndMethod ) {
        
        // Release the content instance
        SCAContentController cc = getFcSCAContentCtrlItf();
        cc.releaseFcContent(content,isEndMethod);
        
        // Pop the request context which has been pushed by pushFcAndGet
        cc.popRequestContext();
    }
    
    
    // -----------------------------------------------------------------------
    // Methods for dealing with Fractal interfaces
    // -----------------------------------------------------------------------

    /**
     * Return the reference of the component context control interface
     * associated with the current component. This method is used by generated
     * subclasses.
     */
    protected ReconfigurableComponentContext getFcCompCtxCtrlItf() {
        if( compctx == null ) {
            compctx =
                getFcInterface(
                    ReconfigurableComponentContext.NAME,
                    ReconfigurableComponentContext.class );
        }
        return compctx;
    }
    
    /**
     * Return the reference of the SCA content control interface associated with
     * the current component.
     */
    private SCAContentController getFcSCAContentCtrlItf() {
        if( cc == null ) {
            cc =
                getFcInterface(
                    SCAContentController.NAME, SCAContentController.class );
        }
        return cc;
    }
    
    /**
     * Return the interface of the specified name and type associated with the
     * current component.
     * 
     * @param name  the interface name
     * @param type  the interface type
     * @throws TinfiRuntimeException  if the interface does not exist
     */
    private <T> T getFcInterface( String name, Class<T> type )
    throws TinfiRuntimeException {
        try {
            T itf = (T) owner.getFcInterface(name);
            return itf;
        }
        catch( NoSuchInterfaceException e ) {
            throw new TinfiRuntimeException(e);
        }
    }
}
