/***
 * OW2 FraSCAti Tinfi
 * Copyright (C) 2008-2009 INRIA, USTL
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: frascati@ow2.org
 *
 * Authors: Lionel Seinturier, Philippe Merle
 */

package org.ow2.frascati.tinfi.control.content;

import java.lang.annotation.Annotation;
import java.lang.reflect.AccessibleObject;
import java.lang.reflect.AnnotatedElement;
import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.Map;

import org.osoa.sca.ComponentContext;
import org.osoa.sca.annotations.Callback;
import org.osoa.sca.annotations.ComponentName;
import org.osoa.sca.annotations.Context;
import org.osoa.sca.annotations.ConversationAttributes;
import org.osoa.sca.annotations.ConversationID;
import org.osoa.sca.annotations.Destroy;
import org.osoa.sca.annotations.EagerInit;
import org.osoa.sca.annotations.Init;
import org.osoa.sca.annotations.Property;
import org.osoa.sca.annotations.Reference;
import org.osoa.sca.annotations.Scope;
import org.ow2.frascati.tinfi.reflect.AnnotatedElementFilter;
import org.ow2.frascati.tinfi.reflect.Filter;
import org.ow2.frascati.tinfi.reflect.Filters;
import org.ow2.frascati.tinfi.reflect.Util;

/**
 * This class stores SCA related metadata for the content class of a component.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 * @author Philippe Merle <Philippe.Merle@inria.fr>
 */
public class ContentClassMetaData {

    // -----------------------------------------------------------------------
    // Factory
    // -----------------------------------------------------------------------
    
    /**
     * Retrieve the SCA related metadata for the specified component class name.
     * 
     * @param fcContentDesc  the component class name (so called content class)
     * @return               the corresponding metadata
     * @throws IllegalContentClassMetaData
     *      if an error occurs while retrieving metadata 
     */
    public static ContentClassMetaData get( String fcContentDesc )
    throws IllegalContentClassMetaData {
        
        ContentClassMetaData ccmd = ccmds.get(fcContentDesc);
        if( ccmd == null ) {
            
            Class<?> c = null;
            try {
                c = Util.loadClass(fcContentDesc);
            }
            catch( ClassNotFoundException cfe ) {
                throw new IllegalContentClassMetaData(cfe);
            }
            
            ccmd = new ContentClassMetaData(c);
            ccmds.put(fcContentDesc,ccmd);
        }
        
        return ccmd;
    }
    
    /**
     * Retrieve the SCA related metadata for the specified component class name.
     * 
     * @param fcContentDesc  the component class name (so called content class)
     * @param cl             the class loader for loading the component class
     * @return               the corresponding metadata
     * @throws IllegalContentClassMetaData
     *      if an error occurs while retrieving metadata 
     */
    public static ContentClassMetaData get( String fcContentDesc, ClassLoader cl )
    throws IllegalContentClassMetaData {
        
        ContentClassMetaData ccmd = ccmds.get(fcContentDesc);
        if( ccmd == null ) {
            
            Class<?> c = null;
            try {
                c = Util.loadClass(fcContentDesc,cl);
            }
            catch( ClassNotFoundException cfe ) {
                throw new IllegalContentClassMetaData(cfe);
            }

            ccmd = new ContentClassMetaData(c);
            ccmds.put(fcContentDesc,ccmd);
        }
        
        return ccmd;
    }
    
    private static Map<String,ContentClassMetaData> ccmds =
        new HashMap<String,ContentClassMetaData>();

    
    // -------------------------------------------------------------------------
    // Public data
    // -------------------------------------------------------------------------
    
    /** The component content class. */
    public Class<?> fcContentClass;
    
    /** The @{@link EagerInit} annotation, if any. */
    public EagerInit eagerinit;

    /** The scope policy, if any. */
    public Scope scope;
    
    /** The conversation attributes, if any. */
    public ConversationAttributes convattr;
    
    /** The maximun age of the conversation in milliseconds. */
    public long convMaxAge = 0;
    
    /** The maximun idle time of the conversation in milliseconds. */
    public long convMaxIdleTime = 0;
    
    /** The constructor annotated with @{@link Constructor}, if any. */
    public Constructor<?> constructorAnnotatedElement;
    
    /** The method annotated with @{@link Init}, if any. */
    public Method initMethod;
    
    /** The method annotated with @{@link Destroy}, if any. */
    public Method destroyMethod;
    
    /** The field or setter method annotated with @{@link Context}, if any. */
    public InjectionPoint<Context> contextAnnotatedElement;
    
    /** The field or setter method annotated with @{@link ComponentName}, if any. */
    public InjectionPoint<ComponentName> componentNameAnnotatedElement;
    
    /** The field or setter method annotated with @{@link ConversationID}, if any. */
    public InjectionPoint<ConversationID> conversationIDAnnotatedElement;
    
    /** List of references. */
    public InjectionPointMap<Reference> refs;

    /** List of properties. */
    public InjectionPointMap<Property> props;
    
    /** The unannotated setter methods. */
    public Map<String,Method> setters;
    
    /** The fields or methods annotated with @{@link Callback}. */
    public AccessibleObject[] callbacks;

    
    // -------------------------------------------------------------------------
    // Constructor to initialize data
    // -------------------------------------------------------------------------
    
    /**
     * Retrieve the SCA related metadata for the specified component class name.
     * 
     * @throws IllegalContentClassMetaData
     *      if an error occurs while retrieving metadata 
     */
    private ContentClassMetaData( Class<?> fcContentClass )
    throws IllegalContentClassMetaData {
        
        this.fcContentClass = fcContentClass;
        
        /*
         * Retrieve the scope policy of the content class.
         */
        scope = fcContentClass.getAnnotation(Scope.class);
        
        /*
         * Retrieve the scope eager init policy of the content class, if any.
         */
        eagerinit = fcContentClass.getAnnotation(EagerInit.class);
        if( eagerinit != null ) {
            if( scope==null || !scope.value().equals("COMPOSITE") ) {
                String msg =
                    "Class "+fcContentClass+
                    " annotated with @EagerInit should be composite-scoped";
                throw new IllegalContentClassMetaData(msg);
            }
        }

        /*
         * Retrieve the conversation attributes of the content class.
         */
        convattr = fcContentClass.getAnnotation(ConversationAttributes.class);
        if( convattr != null ) {
            String ma = convattr.maxAge();
            convMaxAge = convDurationToMilli(ma);
            String mit = convattr.maxIdleTime();
            convMaxIdleTime = convDurationToMilli(mit);
        }
        
        /*
         * Retrieve the constructor annotated with @Constructor, if any.
         * Check that only one such constructor exists.
         */
        Constructor<?>[] constructors =
        	fcContentClass.getDeclaredConstructors();
        Filter<AnnotatedElement> filter =
        	new AnnotatedElementFilter(
    			org.osoa.sca.annotations.Constructor.class);
        Constructor<?>[] ctors = Filters.filter(constructors,filter);
        if( ctors.length > 1 ) {
            String msg =
            	"Only one constructor should be annotated with @Constructor";
            throw new IllegalContentClassMetaData(msg);
        }
        constructorAnnotatedElement = (ctors.length==1) ? ctors[0] : null;

        /*
         * Retrieve the method annotated with @Init, if any.
         * Check that only one such method exists.
         */
        Method[] methods = Util.getAllMethods(fcContentClass);
        filter = new AnnotatedElementFilter(Init.class);
        Method[] inits = Filters.filter(methods,filter);
        initMethod = checkFcUniqueAnnotatedVoidMethod(inits,Init.class);

        /*
         * Retrieve the setter or field annotated with @Destroy, if any.
         * Check that only one such method exists.
         */
        filter = new AnnotatedElementFilter(Destroy.class);
        Method[] destroys = Filters.filter(methods,filter);
        destroyMethod =
            checkFcUniqueAnnotatedVoidMethod(destroys,Destroy.class);

        /*
         * Retrieve the setter or field annotated with @Context, if any.
         * Check that only one such element exists.
         */
        AccessibleObject[] sfs =
            Util.getAllAnnotatedSettersAndFields(fcContentClass,Context.class);
        contextAnnotatedElement =
        	checkSingleton(sfs,Context.class,ComponentContext.class);
        
        /*
         * Retrieve the setter or field annotated with @ComponentName, if any.
         * Check that only one such element exists.
         */
        sfs =
        	Util.getAllAnnotatedSettersAndFields(
        		fcContentClass, ComponentName.class );
        componentNameAnnotatedElement =
        	checkSingleton(sfs,ComponentName.class,String.class);

        /*
         * Retrieve the element annotated with @ConversationID, if any.
         * Check that only one such element exists.
         */
        sfs =
        	Util.getAllAnnotatedSettersAndFields(
    			fcContentClass, ConversationID.class );
        conversationIDAnnotatedElement =
            checkSingleton(sfs,ConversationID.class,Object.class);

        /*
         * Retrieve the elements annotated with @Reference.
         */
        refs = new InjectionPointMap<Reference>(fcContentClass,Reference.class);
        try {
            refs.putAll();
        }
        catch (DuplicationInjectionPointException dipe) {
            throw new IllegalContentClassMetaData(dipe);
        }
        
        /*
         * Retrieve the elements annotated with @Property.
         */
        props = new InjectionPointMap<Property>(fcContentClass,Property.class);
        try {
            props.putAll();
        }
        catch (DuplicationInjectionPointException dipe) {
            throw new IllegalContentClassMetaData(dipe);
        }
        
        /*
         * Retrieve the unannotated setter methods.
         */
        setters = Util.getAllUnAnnotatedSetterMethods(fcContentClass);
        
        /*
         * Retrieve the callbacks declared in the content class.
         */
        callbacks =
        	Util.getAllAnnotatedSettersAndFields(
        		fcContentClass, Callback.class );
    }

    
    // -------------------------------------------------------------------------
    // Implementation specific
    // -------------------------------------------------------------------------

    /**
     * Check that the specified array of setter methods or fields is empty or is
     * a singleton and that the specified type can be injected in this singleton
     * element.
     * 
     * @param <A>        the annotation type associated with the injection point
     * @param aos        the array to be checked
     * @param annotType  the annotation type associated with the injection point
     * @param type       the type to be injected
     * @return           the injection point for the singleton element or
     * 					 <code>null</code> if the array is empty
     * @throws IllegalContentClassMetaData
     * 		if the array is neither empty nor a singleton
     */
    private <A extends Annotation> InjectionPoint<A> checkSingleton(
		AccessibleObject[] aos, Class<A> annotType, Class<?> type )
    throws IllegalContentClassMetaData {
        
        if( aos.length > 1 ) {
            String msg =
                "More than one element annotated with @"+ annotType.getName()+
                " in "+fcContentClass.getName();
            throw new IllegalContentClassMetaData(msg);
        }
        
        if( aos.length == 0 ) {
        	return null;
        }
        
        AccessibleObject ao = aos[0];
        A annot = ao.getAnnotation(annotType);
        InjectionPoint<A> ip = InjectionPointImpl.getInjectionPoint(ao,annot);
        
        Class<?> iptype = ip.getType();
        if( ! type.isAssignableFrom(iptype) ) {
        	String msg =
        		ao.toString()+" should be injectable with "+type.getName();
        	throw new IllegalContentClassMetaData(msg);
        }
        
        return ip;
    }

    /**
     * Check that at most one void returning, no argument, method is annotated
     * with the specified annotation in the given array of methods.
     * 
     * @return  the method or <code>null</code> if there is no such method
     * @throws IllegalContentClassMetaData  if this is not the case
     */
    protected Method checkFcUniqueAnnotatedVoidMethod(
        Method[] methods, Class<? extends Annotation> annot )
    throws IllegalContentClassMetaData {
        
        if( methods.length > 1 ) {
            String msg =
                "More than one element annotated with @"+annot+" in "+
                fcContentClass.getName();
            throw new IllegalContentClassMetaData(msg);
        }        
        
        Method method = (methods.length==1) ? methods[0] : null;
        if( method != null ) {
            if( ! method.getReturnType().equals(void.class) ||
                method.getParameterTypes().length != 0 ) {
                String msg =
                    "The signature of the @"+annot+" annotated method ("+
                    method.getName()+") should be ():void";
                throw new IllegalContentClassMetaData(msg);
            }
        }
        
        return method;
    }
    
    /**
     * Return the duration in milliseconds corresponding to the specified
     * string. The format of the string is defined in the SCA specifications.
     * 
     * @throws IllegalContentClassMetaData  if the string is badly formatted
     */
    private static long convDurationToMilli( String duration )
    throws IllegalContentClassMetaData {
        
        int space = duration.indexOf(' ');
        if( space == -1 ) {
            String msg = "No space in duration string: "+duration;
            throw new IllegalContentClassMetaData(msg);
        }
        
        String numberStr = duration.substring(0,space);
        int number = Integer.parseInt(numberStr);
        if( number < 0 ) {
            String msg = "Negative duration: "+duration;
            throw new IllegalContentClassMetaData(msg);
        }
        
        if( space+1 == duration.length() ) {
            String msg = "No time unit in: "+duration;
            throw new IllegalContentClassMetaData(msg);
        }
        
        String unit = duration.substring(space+1);
        long l = number;
        if( unit.equals("seconds") )       l *= 1000;
        else if( unit.equals("minutes") )  l *= 60*1000;
        else if( unit.equals("hours") )    l *= 60*60*1000;
        else if( unit.equals("days") )     l *= 24*60*60*1000;
        else if( unit.equals("years") )    l *= 365*24*60*60*1000;
        else {
            String msg = "Unsupported time unit in: "+duration;
            throw new IllegalContentClassMetaData(msg);
        }
        
        return l;
    }
}
