/***
 * OW2 FraSCAti Tinfi
 * Copyright (C) 2009 INRIA, USTL
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: frascati@ow2.org
 *
 * Author: Lionel Seinturier
 */

package org.ow2.frascati.tinfi.control.content;

import java.lang.annotation.Annotation;
import java.lang.reflect.AccessibleObject;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.util.HashMap;

import org.ow2.frascati.tinfi.TinfiRuntimeException;
import org.ow2.frascati.tinfi.reflect.SetterMethodFilter;
import org.ow2.frascati.tinfi.reflect.Util;

/**
 * This class manages the injection points of a given type for a specified
 * class. The type is an annotation such as the one for properties or
 * references. The injection points are stored in the inherited map and are
 * indexed by their name. The name is typically a property or a reference name.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 * @since 1.1.2
 */
public class InjectionPointMap<T extends Annotation>
extends HashMap<String,InjectionPoint<T>> {
    
	private static final long serialVersionUID = -2199033442372122424L;

	/** The class. */
    private Class<?> cl;
    
    /** The injection type. */
    private Class<T> annotType;
    
    public InjectionPointMap( Class<?> cl, Class<T> annot ) {
        this.cl = cl;
        this.annotType = annot;
    }
    
    /**
     * Register all injection points by introspecting the class associated with
     * this instance.
     * 
     * @throws DuplicationInjectionPointException
     * 		if for a given name, more than one injection point exist
     */
    public void putAll() throws DuplicationInjectionPointException {
        
        AccessibleObject[] aos =
        	Util.getAllAnnotatedSettersAndFields(cl,annotType);        
        for (AccessibleObject ao : aos) {
            
            /*
             * Check whether the accessible object has a 'name' annotation
             * parameter. If the 'name' annotation parameter is not found, infer
             * the name from the setter method name or the field.
             */
        	String ipname = getInjectionPointName(ao);
            
            /*
             * Check whether the name has already been encountered before. If
             * so, this means that more than one injection point (setter or
             * field) exists, which is unconsistent.
             */
            if( containsKey(ipname) ) {
            	throw new DuplicationInjectionPointException(ipname,cl,annotType);
            }

            T annot = ao.getAnnotation(annotType);
            InjectionPoint<T> ip = InjectionPointImpl.getInjectionPoint(ao,annot);
            put(ipname,ip);
        }        
    }
    
    /**
     * Register the injection point associated with the specified name by
     * introspecting the class associated with this instance. The name is
     * typically a property or a reference name.
     */
    public void put( String name )
    throws NoSuchInjectionPointException, DuplicationInjectionPointException {
        
        AccessibleObject[] aos =
        	Util.getAllAnnotatedSettersAndFields(cl,annotType);        
        for (AccessibleObject ao : aos) {
            
            /*
             * Check whether the accessible object has a 'name' annotation
             * parameter. If the 'name' annotation parameter is not found, infer
             * the name from the setter method name or the field.
             */
        	String ipname = getInjectionPointName(ao);

        	if( ipname.equals(name) ) {
        		
                /*
                 * Check whether the name has already been encountered before.
                 * If so, this means that more than one injection point (setter
                 * or field) exists, which is unconsistent.
                 */
                if( containsKey(name) ) {
                	throw new DuplicationInjectionPointException(name,cl,annotType);
                }

                T annot = ao.getAnnotation(annotType);
                InjectionPoint<T> ip = InjectionPointImpl.getInjectionPoint(ao,annot);
                put(name,ip);
                return;
        	}
        }
        
        throw new NoSuchInjectionPointException(name,cl,annotType);
    }
    
    /**
     * Register the injection point associated with the specified field.
     * 
     * @param field  the field
     * @param annot  the annotation associated with the field
     */
    public void put( Field field, T annot )
    throws DuplicationInjectionPointException {
    	
    	String name = field.getName();
    	
        /*
         * Check whether the name has already been encountered before. If so,
         * this means that more than one injection point exists, which is
         * unconsistent.
         */
        if( containsKey(name) ) {
        	throw new DuplicationInjectionPointException(name,cl,annotType);
        }

    	InjectionPoint<T> ip = InjectionPointImpl.getInjectionPoint(field,annot);
        put(name,ip);
    }
    
    /**
     * Register the injection point associated with the specified setter.
     * 
     * @param setter  the setter
     * @param annot   the annotation associated with the setter
     */
    public void put( Method setter, T annot )
    throws DuplicationInjectionPointException {
    	
    	SetterMethodFilter.checkSetterMethod(setter);
    	String name = SetterMethodFilter.getSetterPropertyName(setter);
    	
        /*
         * Check whether the name has already been encountered before. If so,
         * this means that more than one injection point exists, which is
         * unconsistent.
         */
        if( containsKey(name) ) {
        	throw new DuplicationInjectionPointException(name,cl,annotType);
        }

    	InjectionPoint<T> ip = InjectionPointImpl.getInjectionPoint(setter,annot);
        put(name,ip);
    }
    
    
    // ---------------------------------------------------------------------
    // Implementation specific
    // ---------------------------------------------------------------------
    
    /**
     * Return the name of the injection point associated with the specified
     * element (setter method or field.) Retrieve the name from the {@link
     * #annotType} annotation. If missing, infer the name from the setter method or
     * the field.
     */
    private String getInjectionPointName( AccessibleObject ao ) {
        
    	/*
         * Check whether the accessible object has a 'name' annotation
         * parameter. If the 'name' annotation parameter is not found, infer
         * the name from the setter method name or the field.
         */
    	String name = getNameAnnotationParameter(ao);
        if( name==null || name.length() == 0 ) {
            if( ao instanceof Method ) {
                Method method = (Method) ao;
                name = SetterMethodFilter.getSetterPropertyName(method);
            }
            else {
                Field field = (Field) ao;
                name = field.getName();
            }
        }
        
        return name;
    }
    
    /**
     * If the specified accessible object is annotated with an instance of
     * {@link #annotType} which defines a <code>name</code> parameter of type
     * string, return its value. Else return <code>null</code>.
     */
    private String getNameAnnotationParameter( AccessibleObject ao ) {
        
        try {
            Method meth = annotType.getMethod("name");
            Annotation a = ao.getAnnotation(annotType);
            try {
                Object o = meth.invoke(a);
                if( o instanceof String ) {
                	return (String) o;
                }
            }
            catch( Exception e ) {
            	// Shouldn't occur
                throw new TinfiRuntimeException(e);
            }
        }
        catch( NoSuchMethodException e ) {}
        
        return null;
    }
}
